#define PLUGIN_VERSION "1.6"

#pragma semicolon 1
#pragma newdecls required

#include <sourcemod>
#include <sdktools>
#undef REQUIRE_PLUGIN
#include <l4d_sm_respawn>

#define CVAR_FLAGS	FCVAR_NOTIFY
#define TEAM_SPECTATOR		1
#define TEAM_SURVIVORS 		2
#define TEAM_INFECTED 		3

public Plugin myinfo =
{
	name = "[L4D1 & L4D2] Bebop Addon",
	author = "Dragokas",
	description = "Fixing some glitches of original bebop",
	version = PLUGIN_VERSION,
	url = "https://github.com/dragokas/"
};

/*
	ChangeLog:
	
	1.0 (08-Apr-2021)
	 - First release
	 
	1.1 (19-Apr-2021)
	 - Fixed bug with incorrect detection of spawn state, because players are not connected instantly in OnMapStart().
	 - Added ability to move to Survivors team on "E" (Use) button press.
	 
	1.2 (21-Apr-2021)
	 - Spanwed players detection logic is entirely re-written & simplified.
	 
	1.3 (24-Apr-2021)
	 - Returned SteamId cache logic.
	 - "E" key to move to Survivors team is now working only if there are no other bots to takeover.
	 
	1.4 (03-May-2021)
	 - Added ability to takeover bots with "E" (Use) key.
	 - Added message hint to "Press E" for join Survivor team.
	 - Added command "sm_join" to join survivor team.
	 - Added message hint to player who is waiting for spawn about that they are more tanks on the field when one more tank is died.
	 - Added translation color support.
	 - Fixed some mistakes with key press time restriction.
	 
	1.5 (04-May-2021)
	 - Fixed timer leak.
	 
	1.6 (22-Nov-2021)
	 - Fixed "m_humanSpectatorUserID" error due to checking against non-bot player.
*/

StringMap g_hMapSpawn;
bool g_bRespawnNativeAvailable, g_bLeft4Dead2, g_bHasAwaiters, g_bLateload, g_bMapStarted, g_bSpawned[MAXPLAYERS+1];
Handle g_hSDK_SetHumanSpec, g_hSDK_TakeOverBot;
int g_iTeam[MAXPLAYERS+1], g_UseTime[MAXPLAYERS+1];
char g_sSteamId[MAXPLAYERS+1][32];

public APLRes AskPluginLoad2(Handle myself, bool late, char[] error, int err_max)
{
	EngineVersion test = GetEngineVersion();
	if (test == Engine_Left4Dead2) {
		g_bLeft4Dead2 = true;		
	}
	else if (test != Engine_Left4Dead) {
		strcopy(error, err_max, "Plugin only supports Left 4 Dead 1 & 2.");
		return APLRes_SilentFailure;
	}
	g_bLateload = late;
	return APLRes_Success;
}

public void OnPluginStart()
{
	LoadTranslations("bebop_addon.phrases");
	
	CreateConVar("l4d_bebop_addon_version", PLUGIN_VERSION, "Plugin Version", CVAR_FLAGS | FCVAR_DONTRECORD);
	//AutoExecConfig(true, "l4d_bebop_addon");
	
	HookEvent("player_activate", 		Event_PlayerActivate); // when client becomes fully in-game
	HookEvent("player_team",			Event_PlayerTeam);
	HookEvent("round_end",				Event_RoundEnd);
	HookEvent("player_spawn",			Event_PlayerSpawn);
	
	g_hMapSpawn = new StringMap();
	
	Handle hGameData = LoadGameConfigFile("l4d_respawn_improved");
	if( hGameData == null ) SetFailState("Could not find gamedata file at addons/sourcemod/gamedata/l4d_respawn_improved.txt , you FAILED AT INSTALLING");
	
	StartPrepSDKCall(SDKCall_Player);
	if( PrepSDKCall_SetFromConf(hGameData, SDKConf_Signature, "TakeOverBot") == false )
	{
		LogError("Failed to find signature: CTerrorPlayer::TakeOverBot");
	}
	else {
		PrepSDKCall_AddParameter(SDKType_Bool, SDKPass_Plain);
		g_hSDK_TakeOverBot = EndPrepSDKCall();
		if( g_hSDK_TakeOverBot == null ) LogError("Failed to create SDKCall: CTerrorPlayer::TakeOverBot");
	}
	
	StartPrepSDKCall(SDKCall_Player);
	if( PrepSDKCall_SetFromConf(hGameData, SDKConf_Signature, "SetHumanSpec") == false )
	{
		LogError("Failed to find signature: SurvivorBot::SetHumanSpectator");
	}
	else {
		PrepSDKCall_AddParameter(SDKType_CBasePlayer, SDKPass_Pointer);
		g_hSDK_SetHumanSpec = EndPrepSDKCall();
		if( g_hSDK_SetHumanSpec == null ) LogError("Failed to create SDKCall: SurvivorBot::SetHumanSpectator");
	}
	
	//RegAdminCmd("sm_res", CmdCheck, ADMFLAG_ROOT, "Chack resp allowed");
	
	RegConsoleCmd("sm_join", CmdJoin, "Join survivor team");
	
	if( g_bLateload )
	{
		for( int i = 1; i <= MaxClients; i++ )
		{
			if( IsClientInGame(i) && !IsFakeClient(i) )
			{
				g_bSpawned[i] = true;
				CacheSteamID(i);
				
				g_iTeam[i] = GetClientTeam(i);
				
				if( g_iTeam[i] == TEAM_SPECTATOR )
				{
					HintJoinTeam(i);
				}
			}
		}
		RefreshSpanedList();
	}
}

public Action CmdJoin(int client, int argc)
{
	if( g_UseTime[client] == 0.0 || GetTime() - g_UseTime[client] > 2 )
	{
		if( GetClientTeam(client) == 1 )
		{
			CreateTimer(3.0, Timer_JoinTeam, GetClientUserId(client), TIMER_FLAG_NO_MAPCHANGE);
		}
	}
	g_UseTime[client] = GetTime();
	return Plugin_Handled;
}

public Action Timer_JoinTeam(Handle timer, int UserId)
{
	int client = GetClientOfUserId(client);
	if( client && IsClientInGame(client) )
	{
		if( GetClientTeam(client) == 1 )
		{
			int bot = FindBotToTakeOver(client, TEAM_SURVIVORS);
			if( bot != -1 )
			{
				SDKCall(g_hSDK_SetHumanSpec, bot, client);
				SDKCall(g_hSDK_TakeOverBot, client, true);
			}
			else {
				if( HasAnybodyAtSafeArea() )
				{
					SM_Respawn(client, _, SPAWN_POSITION_SAFEROOM, SPAWN_IN_TEAM_SURVIVOR);
					g_bSpawned[client] = true;
				}
				else {
					ChangeClientTeam(client, TEAM_SURVIVORS);
				}
			}
		}
	}
}

public Action CmdCheck(int client, int argc)
{
	for( int i = 1; i <= MaxClients; i++ )
	{
		if( IsClientInGame(i) && !IsFakeClient(i) )
		{
			ReplyToCommand(client, "%N (team: %i) Spawned? %b. Spawned (ID)? %b", i, GetClientTeam(i), g_bSpawned[i], IsSpawnedSteamId(i));
		}
	}
	return Plugin_Handled;
}

public void OnAllPluginsLoaded()
{
    g_bRespawnNativeAvailable = (GetFeatureStatus(FeatureType_Native, "SM_Respawn") == FeatureStatus_Available);
}

public Action OnPlayerRunCmd(int client, int& buttons, int& impulse, float vel[3], float angles[3])
{
	if( client && g_iTeam[client] == TEAM_SPECTATOR && (buttons & IN_USE) )
	{
		if( g_UseTime[client] == 0.0 || GetTime() - g_UseTime[client] > 2 )
		{
			int bot = FindBotToTakeOver(client, TEAM_SURVIVORS);
			if( bot != -1 )
			{
				SDKCall(g_hSDK_SetHumanSpec, bot, client);
				SDKCall(g_hSDK_TakeOverBot, client, true);
			}
			else {
				ChangeClientTeam(client, TEAM_SURVIVORS);
			}
		}
		g_UseTime[client] = GetTime();
	}
	return Plugin_Continue;
}

public void OnMapStart()
{
	g_bMapStarted = true;
}

public void OnMapEnd()
{
	for( int i = 1; i <= MaxClients; i++ )
	{
		g_bSpawned[i] = false;
	}
	g_bMapStarted = false;
}

void RefreshSpanedList()
{
	g_hMapSpawn.Clear();
	
	for( int i = 1; i <= MaxClients; i++ )
	{
		if( IsClientInGame(i) && !IsFakeClient(i) )
		{
			MarkClientAsSpawned(i);
		}
	}
}

public void OnClientDisconnect_Post(int client)
{
	g_sSteamId[client][0] = '\0';
}

public void OnClientAuthorized(int client, const char[] auth)
{
	if( client )
	{
		CacheSteamID(client);
	}
}

public void Event_RoundEnd(Event event, const char[] name, bool dontBroadcast)
{
	OnMapEnd();
	RefreshSpanedList();
}

public void Event_PlayerSpawn(Event event, const char[] name, bool dontBroadcast)
{
	int client = GetClientOfUserId(event.GetInt("userid"));
	if( client && IsClientInGame(client) )
	{
		g_bSpawned[client] = true;
		MarkClientAsSpawned(client);
	}
}

public void Event_PlayerTeam(Event event, const char[] name, bool dontBroadcast)
{
	int client = GetClientOfUserId(event.GetInt("userid"));
	if( client && IsClientInGame(client) )
	{
		g_iTeam[client] = event.GetInt("team");
		
		if( g_iTeam[client] == TEAM_SPECTATOR && !IsFakeClient(client) )
		{
			HintJoinTeam(client);
		}
	}
}

void HintJoinTeam(int client)
{
	CPrintToChat(client, "%t", "Press_E");
	CreateTimer(15.0, Timer_MsgJoin, GetClientUserId(client), TIMER_REPEAT | TIMER_FLAG_NO_MAPCHANGE);
}

public Action Timer_MsgJoin(Handle timer, int UserId)
{
	int client = GetClientOfUserId(UserId);
	if( client && IsClientInGame(client) )
	{
		if( GetClientTeam(client) == 1 )
		{
			CPrintToChat(client, "%t", "Press_E");
			return Plugin_Continue;
		}
	}
	return Plugin_Handled;
}

public void Event_PlayerActivate(Event event, const char[] name, bool dontBroadcast)
{
	if( !g_bMapStarted )
	{
		return;
	}
	int client = GetClientOfUserId(event.GetInt("userid"));
	if( client && IsClientInGame(client) && !IsFakeClient(client) )
	{
		CreateTimer(7.0, Timer_Respawn, GetClientUserId(client), TIMER_FLAG_NO_MAPCHANGE);
	}
}

public Action Timer_Respawn(Handle timer, int UserId)
{
	int client = GetClientOfUserId(UserId);
	if( client && IsClientInGame(client) && !IsPlayerAlive(client) )
	{
		CheckAndRespawn(client);
	}
}

void CheckAndRespawn(int client) // respawns client if this is his first spawn on this map and if there are no angry tanks on the field
{
	if( !g_bSpawned[client] && !IsSpawnedSteamId(client) )
	{
		if( !HasTanks() )
		{
			if( g_bRespawnNativeAvailable )
			{
				SM_Respawn(client, _, SPAWN_POSITION_ORIGIN);
				g_bSpawned[client] = true;
			}
		}
		else {
			g_bHasAwaiters = true;
			CPrintToChat(client, "\x04%t", "Sorry"); // "Sorry, you can't spawn right now. Need to wait the tank kill.");
		}
	}
}

public void OnClientDisconnect(int client) // spawns all other waiting clients as soon as the last angry tank is died
{
	g_bSpawned[client] = false;

	if( g_bHasAwaiters )
	{
		if( client && IsTank(client) )
		{
			if( !HasTanks(client) )
			{
				WarningMsg();
				CreateTimer(5.0, Timer_RespawnWaiters, _, TIMER_FLAG_NO_MAPCHANGE);
			}
			else {
				MsgMoreTanks();
			}
		}
	}
}

void MsgMoreTanks()
{
	for( int i = 1; i <= MaxClients; i++ )
	{
		if( !g_bSpawned[i] && IsClientInGame(i) && !IsFakeClient(i) && !IsPlayerAlive(i) && GetClientTeam(i) != TEAM_INFECTED && !IsSpawnedSteamId(i) )
		{
			CPrintToChat(i, "%t", "Sorry_More_Tanks"); // "Sorry, you can't spawn right now. There are more tanks left...");
		}
	}
}

void WarningMsg()
{
	for( int i = 1; i <= MaxClients; i++ )
	{
		if( !g_bSpawned[i] && IsClientInGame(i) && !IsFakeClient(i) && !IsPlayerAlive(i) && GetClientTeam(i) != TEAM_INFECTED && !IsSpawnedSteamId(i) )
		{
			CPrintToChat(i, "%t", "Be_Ready"); // "Be ready to respawn...");
		}
	}
}

public Action Timer_RespawnWaiters(Handle timer, int UserId)
{
	g_bHasAwaiters = false;
	
	for( int i = 1; i <= MaxClients; i++ )
	{
		if( !g_bSpawned[i] && IsClientInGame(i) && !IsFakeClient(i) && !IsPlayerAlive(i) && GetClientTeam(i) != TEAM_INFECTED && !IsSpawnedSteamId(i) )
		{
			CheckAndRespawn(i);
		}
	}
}

bool HasTanks(int iExcludeClient = 0)
{
	for( int i = 1; i <= MaxClients; i++ )
	{
		if( i != iExcludeClient && IsTank(i) )
		{
			return true;
		}
	}
	return false;
}

stock bool IsTank(int client)
{
	if( client && client <= MaxClients && IsClientInGame(client) && GetClientTeam(client) == 3 )
	{
		int class = GetEntProp(client, Prop_Send, "m_zombieClass");
		if( class == (g_bLeft4Dead2 ? 8 : 5 ))
			return true;
	}
	return false;
}

stock bool NoBots()
{
	for( int i = 1; i <= MaxClients; i++ )
	{
		if( IsClientInGame(i) && IsFakeClient(i) && GetClientTeam(i) == 2 && IsPlayerAlive(i) )
		{
			return false;
		}
	}
	return true;
}

stock bool StringMap_KeyExists(char[] key)
{
	int v;
	return g_hMapSpawn.GetValue(key, v);
}


bool CacheSteamID(int client)
{
	if( g_sSteamId[client][0] == '\0' )
	{
		if ( !GetClientAuthId(client, AuthId_Steam2, g_sSteamId[client], sizeof(g_sSteamId[])) )
		{
			return false;
		}
		if( strcmp(g_sSteamId[client], "STEAM_ID_STOP_IGNORING_RETVALS") == 0 ) // if Steam system is down
		{
			g_sSteamId[client][0] = '\0';
			return false;
		}
	}
	return true;
}

bool IsSpawnedSteamId(int client)
{
	if( g_bSpawned[client] )
	{
		return true;
	}
	if( !g_sSteamId[client][0] ) // support for clients who re-joined
	{
		return true;
	}
	return StringMap_KeyExists(g_sSteamId[client]);
}

void MarkClientAsSpawned(int client)
{
	g_bSpawned[client] = true;

	if( !g_sSteamId[client][0] )
	{
		CacheSteamID(client);
	}
	if( g_sSteamId[client][0] )
	{
		g_hMapSpawn.SetValue(g_sSteamId[client], 0);
	}
}

stock void CPrintToChat(int iClient, const char[] format, any ...)
{
	static char buffer[192];
	SetGlobalTransTarget(iClient);
	VFormat(buffer, sizeof(buffer), format, 3);
	ReplaceColor(buffer, sizeof(buffer));
	PrintToChat(iClient, "\x01%s", buffer);
}

stock void CPrintToChatAll(const char[] format, any ...)
{
	static char buffer[192];
	for( int i = 1; i <= MaxClients; i++ )
	{
		if( IsClientInGame(i) && !IsFakeClient(i) )
		{
			SetGlobalTransTarget(i);
			VFormat(buffer, sizeof(buffer), format, 2);
			ReplaceColor(buffer, sizeof(buffer));
			PrintToChat(i, "\x01%s", buffer);
		}
	}
}

stock void ReplaceColor(char[] message, int maxLen)
{
	ReplaceString(message, maxLen, "{white}", "\x01", false);
	ReplaceString(message, maxLen, "{cyan}", "\x03", false);
	ReplaceString(message, maxLen, "{orange}", "\x04", false);
	ReplaceString(message, maxLen, "{green}", "\x05", false);
}

int FindBotToTakeOver(int target, int team)
{
	static char sNetClass[16];

	int targetUId = GetClientUserId(target);
	
	if( team == TEAM_SURVIVORS ) // if a survivor bot is already belongs to our target (IDLE)
	{
		for( int i = 1; i <= MaxClients; i++ )
		{
			if( IsClientInGame(i) && GetClientTeam(i) == team && IsFakeClient(i) && IsPlayerAlive(i) )
			{
				GetEntityNetClass(i, sNetClass, sizeof(sNetClass));
				
				if( strcmp(sNetClass, "SurvivorBot") == 0 ) // there are reports, that team + IsFakeClient check are not enough for some (?) reason. What kind of bot is that???
				{
					if( targetUId == GetEntProp(i, Prop_Send, "m_humanSpectatorUserID") )
					{
						return i;
					}
				}
			}
		}
	}
	for( int i = 1; i <= MaxClients; i++ ) // take any other free bot
	{
		if( IsClientInGame(i) && GetClientTeam(i) == team && IsFakeClient(i) && IsPlayerAlive(i) )
		{
			if( team == TEAM_SURVIVORS )
			{
				GetEntityNetClass(i, sNetClass, sizeof(sNetClass));
				
				if( strcmp(sNetClass, "SurvivorBot") == 0 ) // there are reports, that team + IsFakeClient check are not enough for some (?) reason. What kind of bot is that???
				{
					if( 0 == GetEntProp(i, Prop_Send, "m_humanSpectatorUserID")) // forbid takeover other people's bot
					{
						return i;
					}
				}
			}
			else {
				return i;
			}
		}
	}
	return -1;
}

bool HasAnybodyAtSafeArea()
{
	const float AREA_RADIUS = 300.0;

	float vecSafe[3], vecPlayer[3];

	int saferoom = FindEntityByClassname(-1, "info_survivor_position"); // prevents infected from stuck in 0 0 0 position
	if( saferoom != -1 )
	{
		GetEntPropVector(saferoom, Prop_Data, "m_vecOrigin", vecSafe);
		
		for( int i = 1; i <= MaxClients; i++ )
		{
			if( IsClientInGame(i) && GetClientTeam(i) == 2 && IsPlayerAlive(i) )
			{
				GetClientAbsOrigin(i, vecPlayer);
				
				if( GetVectorDistance(vecPlayer, vecSafe) <= AREA_RADIUS )
				{
					return true;
				}
			}
		}
	}
	return false;
}

